import importlib
import os
import shutil
import subprocess

from git import Repo

from autopack.api import PackResponse, get_pack_details
from autopack.errors import AutoPackError, AutoPackInstallationError
from autopack.get_pack import try_get_pack, get_pack
from autopack.pack import Pack
from autopack.utils import (
    find_or_create_autopack_dir,
    load_metadata_file,
    write_metadata_file,
    extract_unique_directory_name,
)


def is_dependency_installed(dependency: str) -> bool:
    try:
        importlib.import_module(dependency)
        return True
    except ImportError:
        return False


def install_dependency(dependency: str, quiet=True):
    try:
        subprocess.check_output(
            ["pip", "install", dependency],
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        if not quiet:
            print(f"{dependency} has been successfully installed.")
    except subprocess.CalledProcessError as e:
        if not quiet:
            print(f"Installation of {dependency} failed with the following error:")
            print(e.output)


def ask_to_install_dependencies(dependencies: list[str], force=False, quiet=True):
    for dependency in dependencies:
        if is_dependency_installed(dependency):
            continue

        if force:
            install_dependency(dependency)
        else:
            if not quiet:
                print(f"This pack requires the dependency {dependency} to be installed. Continue?")
            agree = input("[Yn]")
            if agree.lower() == "y" or agree == "":
                install_dependency(dependency, quiet=quiet)
            elif not quiet:
                print(f"Skipping install of {dependency}")


def install_from_git(pack_data: PackResponse, quiet=True) -> str:
    autopack_dir = find_or_create_autopack_dir()

    url = pack_data.repo_url
    pack_path = os.path.join(autopack_dir, *extract_unique_directory_name(pack_data.repo_url).split("/"))

    if os.path.exists(pack_path):
        if not quiet:
            print("Repo already exists, pulling updates")
        Repo(pack_path).remotes.origin.pull()
    else:
        if not quiet:
            print(f"Cloning repo into {pack_path}")
        Repo.clone_from(url, pack_path)

    return pack_path


def update_metadata_file(pack_id: str, pack_response: PackResponse):
    metadata = load_metadata_file()
    metadata[pack_id] = pack_response.__dict__

    write_metadata_file(metadata)


def install_pack(pack_id: str, force_dependencies=False, quiet=True) -> type[Pack]:
    if not quiet:
        print(f"Installing pack: {pack_id}")

    find_or_create_autopack_dir()

    pack = try_get_pack(pack_id)
    if pack:
        if not quiet:
            print(f"Pack {pack_id} already installed.")
        return pack

    try:
        pack_data = get_pack_details(pack_id, remote=True)

        if not pack_data:
            raise AutoPackInstallationError("Could not find pack details")
    except AutoPackError as e:
        # Maybe do something else
        raise AutoPackInstallationError(f"Could not install pack {e}")
    except BaseException as e:
        raise AutoPackInstallationError(f"Could not install pack {e}")

    try:
        git_dir = install_from_git(pack_data, quiet=quiet)

        update_metadata_file(pack_id, pack_data)
        pack = get_pack(pack_id)

        if pack:
            if pack.dependencies:
                ask_to_install_dependencies(pack.dependencies, force=force_dependencies, quiet=quiet)
            return pack
    except Exception as e:
        raise AutoPackInstallationError(f"Couldn't install pack due to error {e}")

    # Clean up bad repo directories to make sure there aren't bad packs in the .autopack dir
    if git_dir and os.path.isdir(git_dir):
        shutil.rmtree(git_dir)

    raise AutoPackInstallationError("Error: Installation completed but pack could still not be found.")



import importlib
import json
import os
import re
import sys
from asyncio import iscoroutinefunction
from json import JSONDecodeError
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Union, Coroutine

from langchain.chat_models.base import BaseChatModel
from langchain.schema import SystemMessage, BaseMessage
from pydantic import BaseModel

from autopack.errors import AutoPackLoadError
from autopack.pack_response import PackResponse

if TYPE_CHECKING:
    from autopack.pack import Pack


def find_or_create_autopack_dir(depth=0) -> str:
    """Try to find a suitable .autopack directory. Tries in this order:
    1. Directory specified in environment variable AUTOPACK_DIR
    2. Existing .autopack directory in current directory
    3. Existing .autopack directory up to 3 directories up
    4. Creates an .autopack directory in current directory
    """
    env_dir = os.environ.get("AUTOPACK_DIR")
    if env_dir:
        return env_dir

    autopack_dir = os.path.abspath(os.path.join(os.getcwd(), *[os.pardir] * depth, ".autopack"))

    if not os.path.exists(autopack_dir) or not os.path.isdir(autopack_dir):
        if depth > 3:
            os.makedirs(".autopack", exist_ok=True)
            return ".autopack"
        return find_or_create_autopack_dir(depth=depth + 1)

    return autopack_dir


def load_metadata_file() -> dict[str, Any]:
    """Return the parsed contents of the metadata file, returning an empty dict if not found or otherwise failed"""
    metadata_dir = find_or_create_autopack_dir()
    metadata_file = os.path.join(metadata_dir, "pack_metadata.json")

    if not os.path.exists(metadata_file):
        return {}

    with open(metadata_file, "r") as f:
        try:
            return json.load(f)
        except JSONDecodeError:
            return {}


def write_metadata_file(data: dict[str, Any]):
    metadata_dir = find_or_create_autopack_dir()
    metadata_file = os.path.join(metadata_dir, "pack_metadata.json")

    with open(metadata_file, "w+") as f:
        json.dump(data, f)


def find_module(pack_data: PackResponse) -> ModuleType:
    autopack_dir = find_or_create_autopack_dir()
    package_path = pack_data.package_path
    pack_module_path = os.path.join(autopack_dir, extract_unique_directory_name(pack_data.repo_url))

    sys.path.insert(0, autopack_dir)
    sys.path.insert(0, pack_module_path)

    try:
        return importlib.import_module(package_path)
    finally:
        sys.path.remove(autopack_dir)
        sys.path.remove(pack_module_path)


def fetch_pack_object(pack_data: PackResponse) -> type["Pack"]:
    package_path = pack_data.package_path
    class_name = pack_data.class_name
    try:
        module = find_module(pack_data)
        pack_class = getattr(module, class_name)
        return pack_class
    except ImportError:
        raise AutoPackLoadError(
            f"Could not import module '{package_path}'. The path is incorrect or module does not exist."
        )
    except AttributeError:
        raise AutoPackLoadError(
            f"Module '{package_path}' does not contain a class named '{class_name}'. The class name is incorrect."
        )
    except TypeError:
        raise AutoPackLoadError(
            f"'{class_name}' in module '{package_path}' is not a class or cannot be instantiated without arguments."
        )


def format_packs_to_openai_functions(packs: list["Pack"]) -> list[dict[str, Any]]:
    return [format_pack_to_openai_function(pack) for pack in packs]


def format_pack_to_openai_function(pack: "Pack") -> dict[str, Any]:
    # Change this if/when other LLMs support functions
    required = []
    run_args = pack.args
    for arg_name, arg in run_args.items():
        arg_required = arg.pop("required", [])
        if arg_required:
            required.append(arg_name)
        run_args[arg_name] = arg

    return {
        "name": pack.name,
        "description": pack.description,
        "parameters": {"type": "object", "properties": run_args},
        "required": required,
    }


def run_args_from_args_schema(args_schema: type[BaseModel]) -> dict[str, dict[str, str]]:
    run_args: dict[str, Any] = {}
    if not args_schema:
        return run_args

    schema = args_schema.schema()
    if not schema:
        return run_args

    for param_name, param in schema.get("properties", []).items():
        run_args[param_name] = {
            "type": param.get("type", param.get("anyOf", "string")),
            "name": param_name,
            "description": param.get("description", ""),
            "required": param_name in schema.get("required", []),
        }
    return run_args


def functions_bulleted_list(packs: list[type["Pack"]]) -> str:
    functions_string = []
    grouped_packs: dict[str, list[type[Pack]]] = {}
    for pack in packs:
        if not pack.categories:
            continue

        for category in pack.categories:
            if category not in grouped_packs:
                grouped_packs[category] = []
            grouped_packs[category].append(pack)

    for category, category_packs in grouped_packs.items():
        functions_string.append(f"\n## {category}")
        sorted_by_name = sorted(category_packs, key=lambda p: p.name)
        for pack in sorted_by_name:
            args: list[dict[str, str]] = []
            if pack.args_schema:
                args = list(run_args_from_args_schema(pack.args_schema).values())
            args_signature = ", ".join([f"{arg.get('name')}: {arg.get('type')}" for arg in args])
            args_descriptions = (
                "; ".join([f"{arg.get('name')} ({arg.get('type')}): {arg.get('description')}" for arg in args])
                or "None."
            )
            functions_string.append(
                f"- {pack.name}({args_signature}): {pack.description} | Arguments: {args_descriptions}"
            )

    return "\n".join(functions_string)


def functions_summary(packs: list["Pack"]) -> str:
    return ", ".join([f"{pack.name}" for pack in packs])


def extract_unique_directory_name(repo_url: str) -> str:
    repo_name = repo_url.split("/")[-1].replace(".git", "")
    # Replace any non-alphanumeric characters with underscores
    return re.sub(r"[^a-zA-Z0-9]", "_", repo_name)


def call_llm(prompt: str, llm: Union[BaseChatModel, Callable[[str], str]]) -> str:
    """
    Call the given LLM  with the specified prompt.

    The function supports both callable LLMs and LLMs that are instances of the BaseChatModel class.

    Args:
        prompt (str): The prompt to feed to the LLM.
        llm (Union[BaseChatModel, Callable]): The LLM to call.

    Returns:
        str: The response from the LLM.
    """
    if isinstance(llm, BaseChatModel):
        message = SystemMessage(content=prompt)
        response = llm(messages=[message])
        if isinstance(response, BaseMessage):
            return response.content
        else:
            return response
    elif callable(llm):
        return llm(prompt)

    return ""


async def acall_llm(prompt: str, llm: Union[BaseChatModel, Callable[[str], str], Coroutine[Any, Any, str]]) -> str:
    """
    Asynchronously call the given LLM  with the specified prompt.

    The function supports both callable LLMs and LLMs that are instances of the BaseChatModel class.

    Args:
        prompt (str): The prompt to feed to the LLM.
        llm (Union[BaseChatModel, Awaitable[Callable]]): The LLM to call.

    Returns:
        str: The response from the LLM.
    """
    if isinstance(llm, BaseChatModel):
        message = SystemMessage(content=prompt)
        response = await llm._call_async(messages=[message])
        if isinstance(response, BaseMessage):
            return response.content
        else:
            return response
    elif callable(llm) and iscoroutinefunction(llm):
        return await llm(prompt)

    return call_llm(prompt, llm)  # type: ignore



from dataclasses import dataclass

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class PackResponse:
    """Class to store metadata about a (possibly uninstalled) pack"""

    pack_id: str
    package_path: str
    class_name: str
    repo_url: str



import json
import os
from json import JSONDecodeError
from urllib.parse import urljoin

import requests
from marshmallow import ValidationError

from autopack.errors import AutoPackFetchError
from autopack.pack_response import PackResponse
from autopack.pack_search_response import PackSearchResponse
from autopack.utils import find_or_create_autopack_dir

API_URL = os.environ.get("AUTOPACK_API_URL", "https://autopack.ai/")


def get_pack_details(pack_id: str, remote=False) -> PackResponse:
    if remote:
        return get_pack_details_remotely(pack_id)

    return get_pack_details_locally(pack_id)


def get_pack_details_locally(pack_id: str) -> PackResponse:
    metadata_dir = find_or_create_autopack_dir()

    metadata_file = os.path.join(metadata_dir, "pack_metadata.json")

    if not os.path.exists(metadata_file):
        raise AutoPackFetchError(f"Metadata file does not exist, please install or re-install {pack_id}.")

    with open(metadata_file, "r") as f:
        try:
            metadata = json.load(f)
        except JSONDecodeError as e:
            raise AutoPackFetchError(f"Could  can't fetch locally. Please install or re-install {pack_id}. {e}")

    pack_metadata = metadata.get(pack_id)
    if not pack_metadata:
        raise AutoPackFetchError(f"Could  can't find pack locally. Please install {pack_id}")

    return PackResponse(**pack_metadata)


def get_pack_details_remotely(pack_id: str) -> PackResponse:
    endpoint = "/api/details"

    url = urljoin(API_URL, endpoint)
    params = {"id": pack_id}

    response = requests.get(url, params=params)
    if response.status_code == 200:
        data = response.json()

        try:
            return PackResponse(**data)
        except (ValidationError, TypeError) as e:
            message = f"Pack fetch received invalid data: {e}"
            raise AutoPackFetchError(message)

    elif response.status_code <= 500:
        raise AutoPackFetchError(f"Error: {response.status_code}")
    else:
        raise AutoPackFetchError(f"Error: {response.status_code}")


def pack_search(query: str) -> list[PackSearchResponse]:
    endpoint = "/api/search"
    url = urljoin(API_URL, endpoint)
    params = {"query": query}

    response = requests.get(url, params=params)
    if response.status_code == 200:
        data = response.json()

        try:
            return [PackSearchResponse(**datum) for datum in data["packs"]]
        except (ValidationError, TypeError) as e:
            message = f"Pack fetch received invalid data: {e}"
            print(message)
            raise AutoPackFetchError(message)

    elif response.status_code <= 500:
        print(f"Error: {response.status_code}")
        return []
    else:
        print(f"Error: {response.status_code}")
        error_message = f"Error: {response.status_code}"
        raise AutoPackFetchError(error_message)



TOOL_SELECTION_TEMPLATE = """As the AI Tool Selector your responsibility is to identify functions (tools) that could be useful for an autonomous AI agent to accomplish a given task.

Analyze the task and available functions, and determine which functions could be useful. Consider functions that can achieve the goal directly or indirectly, in combination with other tools.

When a task involves coding, prioritize tools that can execute code. Also, keep an eye on functions that, even if not explicitly coding-related, may indirectly aid in achieving the task objectives when combined with other tools.

Only recommend programming functions if the task explicitly requires programming.

Task:
Your original task, given by the human, is:
{task}

Available functions:
You may only recommend functions from the following list:
{functions}.

Please respond with a comma-separated list of function names, excluding parentheses and arguments. Do not include any other explanatory text.

By providing more flexibility in the selection and emphasizing the consideration of alternative functions, we can ensure a wider range of function recommendations that align with the given task.
"""

GET_MORE_TOOLS_TEMPLATE = """As the AI Tool Selector your responsibility is to identify functions (tools) that could be useful for an autonomous AI agent to accomplish a given task.

Analyze the functions request and determine which functions could be useful. Consider functions that can achieve the goal directly or indirectly, in combination with other tools.

Only recommend programming functions if the task explicitly requires programming.

Functions request:
The Autonomous AI has made this request for more tools: {functions_request}

Task:
Your original task, given by the human, is:
{task}

Available functions:
You may only recommend functions from the following list:
{functions}.

Please respond with a comma-separated list of function names, excluding parentheses and arguments. Do not include any other explanatory text.

By providing more flexibility in the selection and emphasizing the consideration of alternative functions, we can ensure a wider range of function recommendations that align with the given task.
"""



from langchain.tools import BaseTool
from pydantic import Field

from autopack import Pack


class LangchainWrapper(BaseTool):
    """Thin wrapper around a Pack to allow it to be a LangChain-compatible Tool"""

    pack: Pack = Field(..., description="The AutoPack Pack")

    def __init__(self, **kwargs):
        pack = kwargs.get("pack")
        kwargs["name"] = pack.name
        kwargs["description"] = pack.description
        kwargs["args_schema"] = pack.args_schema
        super().__init__(**kwargs)

    def _run(self, *args, **kwargs):
        return self.pack.run(*args, **kwargs)

    async def _arun(self, *args, **kwargs):
        return self.pack.arun(*args, **kwargs)

    def is_single_input(self) -> bool:
        return False



from dataclasses import dataclass
from typing import Any

from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class PackSearchResponse:
    """Class for display of search results"""

    pack_id: str
    package_path: str
    class_name: str
    repo_url: str
    name: str
    description: str
    dependencies: str
    run_args: dict[str, Any]
    categories: list[str]



import argparse

from autopack.installation import install_pack
from autopack.search import print_search


def parse_args():
    parser = argparse.ArgumentParser(description="AutoPack CLI tool")

    subparsers = parser.add_subparsers(dest="command")
    install_parser = subparsers.add_parser("install", help="Install a pack")
    install_parser.add_argument("pack", help="ID of the pack to install")

    search_parser = subparsers.add_parser("search", help="Search for packs")
    search_parser.add_argument("query", help="The search query")

    parser.add_argument(
        "-f",
        "--force",
        help="Force automatic dependency installation",
        action="store_true",
    )

    return parser.parse_args()


def main():
    args = parse_args()

    if args.command == "install":
        result = install_pack(args.pack, args.force, quiet=False)
        if result:
            print("Installation completed")
        else:
            print("Installation failed")

    if args.command == "search":
        print_search(args.query)


if __name__ == "__main__":
    main()
